package onig4j;

import java.io.FileInputStream;
import java.io.IOException;
import java.nio.charset.Charset;
import java.util.EnumSet;
import java.util.Properties;
import java.util.logging.Logger;
import onig4j.OnigRegex.Callback;
import org.junit.Test;
import static org.junit.Assert.*;

/**
 *
 * @author calico
 */
public class OnigRegexTest {

    @Test
    public void newOnigRegexDefault() {
        OnigRegex regex = new OnigRegex("a(.*)b|[e-f]+");
        regex.close();
    }

    @Test
    public void newOnigRegexJava() {
        OnigRegex regex = new OnigRegex("\\p{Lu}", OnigSyntaxType.ONIG_SYNTAX_JAVA);
        regex.close();
    }

    @Test(expected = java.util.regex.PatternSyntaxException.class)
    public void newOnigRegexThrowPatternSyntaxException() {
        try {
            OnigRegex regex = new OnigRegex("\\p{InGreek}");
            regex.close();
            
        } catch (RuntimeException ex) {
            System.out.println(ex.getMessage());
            assertNotNull(ex.getMessage());
            throw ex;
        }
    }
    
    @Test
    public void getBackRefNumber() {
        OnigRegex regex = new OnigRegex("(?<foo>a*)(?<bar>b*)(?<foo>c*)");
        
        int num = regex.getBackRefNumber("foo", null);
        assertEquals(3, num);

        num = regex.getBackRefNumber("bar", null);
        assertEquals(2, num);

        regex.close();
    }
    
    @Test(expected = java.lang.IllegalArgumentException.class)
    public void getBackRefNumberThrowIllegalArgumentException() {
        OnigRegex regex = new OnigRegex("(?<foo>a*)(?<bar>b*)(?<foo>c*)");
        
        int num = regex.getBackRefNumber("buz", null);

        regex.close();
    }

    @Test
    public void getGroupNumbers() {
        OnigRegex regex = new OnigRegex("(?<foo>a*)(?<bar>b*)(?<foo>c*)");
        
        int[] list = regex.getGroupNumbers("foo");
        assertEquals(2, list.length);
        assertEquals(1, list[0]);
        assertEquals(3, list[1]);

        list = regex.getGroupNumbers("bar");
        assertEquals(1, list.length);
        assertEquals(2, list[0]);

        list = regex.getGroupNumbers("buz");
        assertEquals(0, list.length);
        
        regex.close();
    }
    
    @Test
    public void getEncoding() {
        final String pattern = "a(.*)b|[e-f]+";
        OnigRegex regex = new OnigRegex(pattern);
        assertEquals(Charset.forName("UTF-16LE"), regex.getEncoding());
        regex.close();
    }
    
    @Test
    public void getOptionType() {
        final String pattern = "a(.*)b|[e-f]+";
        OnigRegex regex = null;
        OnigOptionType option = OnigOptionType.ONIG_OPTION_NONE;
        EnumSet<OnigOptionType> options = null;
        
        regex = new OnigRegex(pattern);
        options = regex.getOptionType();
        assertEquals(1, options.size());
        assertTrue(options.contains(option));
        regex.close();

        option = OnigOptionType.ONIG_OPTION_EXTEND;
        regex = new OnigRegex(pattern, option);
        options = regex.getOptionType();
        assertEquals(1, options.size());
        assertTrue(options.contains(option));
        regex.close();

        option = OnigOptionType.ONIG_OPTION_MULTILINE;
        regex = new OnigRegex(pattern, option);
        options = regex.getOptionType();
        assertEquals(1, options.size());
        assertTrue(options.contains(option));
        regex.close();
    }
    
    @Test
    public void getCaseFoldType() {
        final String pattern = "a(.*)b|[e-f]+";
        OnigRegex regex = null;
        
        regex = new OnigRegex(pattern);
        assertEquals(OnigCaseFoldType.ONIGENC_CASE_FOLD_MIN, regex.getCaseFoldType());
        regex.close();
    }
    
    @Test
    public void getSyntaxType() {
        final String pattern = "a(.*)b|[e-f]+";
        OnigSyntaxType syntax = null;
        OnigRegex regex = null;
        
        regex = new OnigRegex(pattern);
        syntax = regex.getSyntaxType();
        assertEquals(OnigSyntaxType.ONIG_SYNTAX_DEFAULT, syntax);
        
        regex = new OnigRegex(pattern, OnigSyntaxType.ONIG_SYNTAX_JAVA);
        syntax = regex.getSyntaxType();
        
        assertEquals(OnigSyntaxType.ONIG_SYNTAX_JAVA, syntax);
        
        regex.close();
    }
    
    @Test
    public void search() {
        OnigRegex regex = new OnigRegex("\\p{Lu}", OnigSyntaxType.ONIG_SYNTAX_JAVA);
        OnigRegion region = new OnigRegion();
        int ret = OnigRegex.ONIG_MISMATCH;
        
        System.out.println("node is " + region.getCaptureTreeNode());
        
//        ret = regex.search("abcdEfg", region, (1 << 10));
        ret = regex.search("abcdEfg", region);
        assertEquals(4, ret);

        System.out.printf("allocated is %d\n", region.allocated());
        System.out.printf("count is %d\n", region.count());
        
        System.out.printf("onig_number_of_capture_histories is %d\n", regex.captureHistoryCount());
        if (regex.captureHistoryCount() > 0) {
            OnigCaptureTreeNode root = region.getCaptureTreeNode();
        }
        
        System.out.printf("match at %d\n", ret);
        for (int i = 0; i < region.count(); i++) {
            System.out.printf("%d: (%d-%d)\n", i, region.begin(i), region.end(i));
        }
        
        ret = regex.search("abcdefg", region);
        assertEquals(OnigRegex.ONIG_MISMATCH, ret);
        
        regex.close();
        region.close();
    }

    @Test
    public void match() {
        OnigRegex regex = new OnigRegex("a(.*)b|[e-f]+");
        OnigRegion region = new OnigRegion();
        int ret = OnigRegex.ONIG_MISMATCH;

        ret = regex.search("zzzzaffffffffb", region);
        assertEquals(4, ret);
        
        ret = regex.match("zzzzaffffffffb", 4, region);
        assertEquals(10, ret);

        System.out.printf("allocated at %d\n", region.allocated());
        System.out.printf("match bytes is %d\n", ret);
        for (int i = 0; i < region.count(); i++) {
            System.out.printf("%d: (%d-%d)\n", i, region.begin(i), region.end(i));
        }
        
        regex.close();
        region.close();
    }
    
    @Test
    public void getVersion() throws IOException {
        final Properties prop = new Properties();
        prop.load(new FileInputStream("nbproject/project.properties"));
        assertEquals(prop.getProperty("product.version"), OnigRegex.getVersion());
    }
    
    @Test
    public void setWarningListener() {
        OnigRegex.clearVerboseWarningListener();
        final OnigRegex.WarningListener listener
                = new OnigRegex.WarningListener() {
                        public void warning(String msg) {
                            Logger.getLogger(OnigRegex.WarningListener.class.getName()).warning(msg);
                            assertNotNull(msg);
                        }
                    };
        assertTrue(OnigRegex.setWarningListener(listener));
        OnigRegex regex = new OnigRegex("あいうえお]");
        regex.close();
    }
    
    @Test
    public void setVerboseWarningListener() {
        OnigRegex.clearWarningListener();
        final OnigRegex.WarningListener listener
                = new OnigRegex.WarningListener() {
                        public void warning(String msg) {
                            Logger.getLogger(OnigRegex.WarningListener.class.getName()).severe(msg);
                            assertNotNull(msg);
                        }
                    };
        assertTrue(OnigRegex.setVerboseWarningListener(listener));
        OnigRegex regex = new OnigRegex("(?:.?)?");
        regex.close();
    }
    
    @Test
    public void onig_new_deluxe() {
        OnigRegex regex = new OnigRegex("[あ-お]", OnigCaseFoldType.ONIGENC_CASE_FOLD_MIN, OnigOptionType.ONIG_OPTION_IGNORECASE);
        
        assertEquals(2, regex.search("abうdefg"));
        
        regex.close();
    }
    
    @Test
    public void captureHistories() {
        OnigRegex regex = new OnigRegex("a(.*)b|[e-f]+");
        OnigRegion region = new OnigRegion();
        int ret = OnigRegex.ONIG_MISMATCH;
        OnigCaptureTreeNode capture = null;
        
        ret = regex.search("abcdefg", region);
        capture = region.getCaptureTreeNode();
        System.out.println("node is " + capture);
        assertNull(capture);
        
        OnigSyntaxType syntax = OnigSyntaxType.ONIG_SYNTAX_DEFAULT.clone();
        syntax.setOperator2(syntax.getOperator2() | OnigSyntaxType.ONIG_SYN_OP2_ATMARK_CAPTURE_HISTORY);
        
        regex = new OnigRegex("(?@a)*", syntax);
//        ret = regex.search("abcdefg", region, (1 << 10));
        ret = regex.search("abcdefg", region);

        capture = region.getCaptureTreeNode();
        System.out.println("node is " + capture);
        assertNotNull(capture);
        OnigCaptureTreeNode[] children = capture.getChildNodes();
        assertNotNull(children);
        
        System.out.printf("allocated is %d\n", region.allocated());
        System.out.printf("count is %d\n", region.count());
        
        System.out.printf("onig_number_of_capture_histories is %d\n", regex.captureHistoryCount());
        if (regex.captureHistoryCount() > 0) {
            OnigCaptureTreeNode root = region.getCaptureTreeNode();
        }
        
        System.out.printf("match at %d\n", ret);
        for (int i = 0; i < region.count(); i++) {
            System.out.printf("%d: (%d-%d)\n", i, region.begin(i), region.end(i));
        }
        
        ret = regex.search("abcdefg", region);
        assertEquals(0, ret);
        
        regex.close();
        region.close();
        syntax.close();
    }
    
    @Test
    public void nameCount() {
        OnigRegex regex = new OnigRegex("a(.*)b|[e-f]+");

        assertEquals(0, regex.nameCount());        
        regex.close();
        
        regex = new OnigRegex("(?<hoge>.*)+");

        assertEquals(1, regex.nameCount());
        
        regex.close();
    }
    
    @Test
    public void foreachName() {
        final OnigRegex regex = new OnigRegex("(?<hogｅ>.*)+(?<ふご>\\h)?(?<hogｅ>.*)?");

        assertEquals(2, regex.nameCount());
        
        final Callback callbakc
                = new OnigRegex.Callback() {
                        public int call(String name, int[] groups, OnigRegex reg) {
                            assertEquals(regex, reg);
                            
                            if ("ふご".equals(name)) {
                                assertEquals(1, groups.length);
                                assertEquals(2, groups[0]);
                                
                            } else {
                                assertEquals("hogｅ", name);
                                assertEquals(2, groups.length);
                                assertEquals(1, groups[0]);
                                assertEquals(3, groups[1]);
                            }
                            return 0;
                        }
                    };
                
        int ret = regex.foreachName(callbakc);
        assertEquals(0, ret);
        
        regex.close();
    }
}